Attribute VB_Name = "Module2"
'
' Notes on subclassing with Comctl32.DLL:
'
'   1.  "Each subclass is uniquely identified by the address of the pfnSubclass and its uIdSubclass"
'       (quote from Microsoft.com).
'
'   2.  For a particular hWnd, the last procedure hooked will be the first to execute.
'       In other words, it's a LIFO (last-in-first-out, or last-hooked-first-to-execute) system.
'
'   3.  If we call SetWindowSubclass repeatedly with the same hWnd, same pfnSubclass,
'       same uIdSubclass, and same dwRefData, it does nothing at all.
'       Not even the order of the hooked functions will change,
'       even if other functions were hooked later, and then it was called again
'       with the same hWnd, pfnSubclass, uIdSubclass, and dwRefData.
'
'   4.  Similar to the above, if we call SetWindowSubclass repeatedly,
'       and nothing changes but the dwRefData, the dwRefData is changed like we want,
'       but the order of execution of the functions still stays the same as it was.
'        "To change reference data you can make subsequent calls to SetWindowSubclass"
'       (quote from Microsoft.com).
'
'   5.  When unhooking, we can call RemoveWindowSubclass in any order we like, with no harm.
'
'   6.  We don't have to call DefSubclassProc in a particular hooked function, but if we don't,
'       all other "downstream" hooked functions won't execute.
'
'   7.  If "uMsg = WM_DESTROY" we should absolutely call DefSubclassProc so that other
'       possible hooks can also unhook.
'
'   8.  A hooked function will get executed even AFTER the IDE "Stop" button is pressed.
'       This gives us an opportunity to unhook everything if things are done correctly.
'       Note that all COM objects are un-instantiated BEFORE the hooked functions are
'       called when "uMsg = WM_DESTROY" when the "Stop" button is pressed.
'
'   9.  dwRefData can be used for whatever we want.
'
'
' The following is a stub that can be used for creating new functions to hook:
'
'        Private Function MySpecialHook(ByVal hWnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long, ByVal uIdSubclass As Long, ByVal dwRefData As Long) As Long
'            If uMsg = WM_DESTROY Then
'                Call RemoveWindowSubclass(hWnd, AddressOf_MySpecialHook, hWnd)
'                MySpecialHook = DefSubclassProc(hWnd, uMsg, wParam, lParam)
'                Exit Function
'            End If
'            '
'            ' Do what we need to do...
'
'
'            ' If the following line is called, other hooks will also execute.
'            ' It's not absolutely necessary that we do this (except for above when uMsg = WM_DESTROY).
'            MySpecialHook = DefSubclassProc(hWnd, uMsg, wParam, lParam)
'        End Function
'
' ... and also ...
'
'       Private Function AddressOf_MySpecialHook() as long
'           AddressOf_MySpecialHook = ProcedureAddress(AddressOf MySpecialHook)
'       End Function
'
'
' To use all of this, do the following steps:
'
'   1.  Write our MySpecialHook function, naming it what we like.
'
'   2.  Write a AddressOf_MySpecialHook function that gives us the address of our MySpecialHook function.
'       The ProcedureAddress function will help with this.
'
'   3.  Make a call to SubclassSomeWindow, and that's it.  It will automatically unhook if done correctly.
'       And the IDE "Stop" button should be okay too.
'
Option Explicit
'
Private Const WM_DESTROY = &H2&
'
Private Declare Function SetWindowSubclass Lib "comctl32.dll" Alias "#410" (ByVal hWnd As Long, ByVal pfnSubclass As Long, ByVal uIdSubclass As Long, Optional ByVal dwRefData As Long) As Long
Private Declare Function GetWindowSubclass Lib "comctl32.dll" Alias "#411" (ByVal hWnd As Long, ByVal pfnSubclass As Long, ByVal uIdSubclass As Long, pdwRefData As Long) As Long
Private Declare Function DefSubclassProc Lib "comctl32.dll" Alias "#413" (ByVal hWnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Private Declare Function RemoveWindowSubclass Lib "comctl32.dll" Alias "#412" (ByVal hWnd As Long, ByVal pfnSubclass As Long, ByVal uIdSubclass As Long) As Long
'
'
'**************************************************************************************
' The following MODULE level stuff is specific to individual subclassing needs.
'**************************************************************************************
'
Private FixedSize_hWndSaved As New Collection
Private FixedSize_StartupWidth As New Collection
Private FixedSize_StartupHeight As New Collection
Private Type MINMAXINFO
    ptReserved As POINTAPI
    ptMaxSize As POINTAPI
    ptMaxPosition As POINTAPI
    ptMinTrackSize As POINTAPI
    ptMaxTrackSize As POINTAPI
End Type
'

Public Sub SubclassSomeWindow(hWnd As Long, AddressOf_ProcToHook As Long, Optional dwRefData As Long)
    Call SetWindowSubclass(hWnd, AddressOf_ProcToHook, hWnd, dwRefData)
End Sub

Public Function GetSubclassRefData(hWnd As Long, AddressOf_ProcToHook As Long) As Long
    ' This one is used only to fetch the optional dwRefData you may have specified when calling SubclassSomeWindow.
    ' Typically this would only be used by the hooked procedure, but it is available to anyone.
    Call GetWindowSubclass(hWnd, AddressOf_ProcToHook, hWnd, GetSubclassRefData)
End Function

Public Sub UnSubclassSomeWindow(hWnd As Long, AddressOf_ProcToHook As Long)
    ' Only needed if we specifically want to un-subclass before we're closing the form (or control),
    '   otherwise, it's automatically taken care of when the window closes.
    ' Be careful, some subclassing may require additional cleanup that's not done here.
    Call RemoveWindowSubclass(hWnd, AddressOf_ProcToHook, hWnd)
End Sub

Private Function ProcedureAddress(AddressOf_TheProc As Long)
    ' A private "helper" function for writing the AddressOf_... functions (see above notes).
    ProcedureAddress = AddressOf_TheProc
End Function

'**************************************************************************************
'**************************************************************************************
'**************************************************************************************
'
' The following are our functions to be hooked, along with their AddressOf_... function.
' All of the following should be Private to make sure we don't accidentally call it.
' But the actual hooked proc can't be private because we'll be using the AddressOf on it.
'
'**************************************************************************************
'**************************************************************************************
'**************************************************************************************

Public Function FixedSize_Proc(ByVal hWnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long, ByVal uIdSubclass As Long, ByVal dwRefData As Long) As Long
    If uMsg = WM_DESTROY Then
        Call RemoveWindowSubclass(hWnd, AddressOf_FixedSize_Proc, hWnd)
        FixedSize_Proc = DefSubclassProc(hWnd, uMsg, wParam, lParam)
        '
        ' We've got a bit of extra clean-up on this one.
        ' This is the kind of place where we have to be careful to protect the IDE and using "Stop".  Collections will be gone.
        If FixedSize_FormIsOnList(hWnd) Then ' Make sure we don't try and do it more than once.  Also, this protects the IDE on "Stop".
            FixedSize_hWndSaved.Remove Format$(hWnd)
            FixedSize_StartupWidth.Remove Format$(hWnd)
            FixedSize_StartupHeight.Remove Format$(hWnd)
        End If
        '
        Exit Function
    End If
    '
    ' On this one, we use dwRefData to let us know it's the first time through so we can do some setup.
    '
    ' Called as follows:
    '       SubclassSomeWindow Me.hWnd, AddressOf FixedSize_Proc, FixedSize_RefData(Me)
    ' We may also want call it like this, just for additional IDE protection:
    '       If (frm.Height > ScreenHeight) Or (frm.Width > ScreenWidth) Or (IsCompiled) Then SubclassSomeWindow Me.hWnd, AddressOf FixedSize_Proc, FixedSize_RefData(Me)
    '
    ' NOTE: If done in the form LOAD event, the form will NOT have been resized from a smaller monitor.
    '       If done in form ACTIVATE or anywhere else, we're too late, and the form will have been resized.
    '
    ' ALSO: If you're in the IDE, and the monitors aren't big enough, do NOT open the form in design mode.
    '       So long as you don't open it, everything is fine, although you can NOT compile in the IDE.
    '       If you're compiling without large enough monitors, you MUST do a command line compile.
    '
    ' This can simultaneously be used by as many forms as will need it.
    '
    ' NOTICE:  Be sure the window is moved (possibly centered) AFTER this is call, or we may not see WM_GETMINMAXINFO until a bit later.
    '
    Dim PelWidth As Long
    Dim PelHeight As Long
    Dim MMI As MINMAXINFO
    Const WM_GETMINMAXINFO As Long = &H24
    '
    If dwRefData Then ' Do our setup.
        PelWidth = dwRefData And &HFFFF&
        PelHeight = (dwRefData And &H7FFF0000) \ &H10000
        '
        ' Destroy dwRefData so we don't do this multiple times.
        dwRefData = 0
        Call SubclassSomeWindow(hWnd, AddressOf_FixedSize_Proc, dwRefData)
        '
        ' Save the original form size.
        If Not FixedSize_FormIsOnList(hWnd) Then ' Make sure it hasn't already been done.
            FixedSize_hWndSaved.Add hWnd, Format$(hWnd)
            FixedSize_StartupWidth.Add PelWidth, Format$(hWnd)
            FixedSize_StartupHeight.Add PelHeight, Format$(hWnd)
        End If
    End If
    '
    ' And now we force our size to not change.
    '
    On Error Resume Next ' This will protect us from form somehow not being on the list.
        If uMsg = WM_GETMINMAXINFO Then
            ' Force the form to stay at initial size.
            PelWidth = FixedSize_StartupWidth.Item(Format$(hWnd))
            PelHeight = FixedSize_StartupHeight.Item(Format$(hWnd))
            '
            CopyMemory MMI, ByVal lParam, LenB(MMI)
            '
            MMI.ptMinTrackSize.x = PelWidth
            MMI.ptMinTrackSize.y = PelHeight
            MMI.ptMaxTrackSize.x = PelWidth
            MMI.ptMaxTrackSize.y = PelHeight
            '
            CopyMemory ByVal lParam, MMI, LenB(MMI)
            Exit Function ' If we process the message, we must return 0 and not let more hook code execute.
        End If
    On Error GoTo 0
    '
    ' Give control to other hooks, if they exist.
    FixedSize_Proc = DefSubclassProc(hWnd, uMsg, wParam, lParam)
End Function

Public Function FixedSize_RefData(frm As VB.Form) As Long
    ' We must use this to pass the form's initial width and height.
    ' Note that using GetWindowRect absolutely doesn't work.  It reports an already resized value.
    '
    Dim PelWidth As Long
    Dim PelHeight As Long
    '
    PelWidth = frm.Width \ TwipsPerPixelX
    PelHeight = frm.Height \ TwipsPerPixelY
    '
    ' Push PelHeight to high two-bytes, and add PelWidth.
    ' This will easily accomodate any monitor in the foreseeable future.
    FixedSize_RefData = PelHeight * &H10000 + PelWidth
End Function

Private Function FixedSize_FormIsOnList(hWnd As Long) As Boolean
    ' Used only internally to determine if a form has been subclassed.
    Dim l As Long
    '
    On Error Resume Next
        l = FixedSize_hWndSaved.Item(Format$(hWnd)) ' This will cause error and leave l=0 if no item in collection.
    On Error GoTo 0
    FixedSize_FormIsOnList = l <> 0
End Function

Private Function AddressOf_FixedSize_Proc() As Long
    AddressOf_FixedSize_Proc = ProcedureAddress(AddressOf FixedSize_Proc)
End Function
